"""Setting up models of controllers including their registers"""

from flask import Blueprint
from flask import flash
from flask import redirect
from flask import render_template
from flask import request
from flask import url_for
from werkzeug.exceptions import abort
import data

bp = Blueprint('controller_types', __name__)


@bp.route('/types')
def list():
    models = data.get_controller_types()
    return render_template('controller_types/list.html',
                           config=models,
                           footer=data.get_info())


@bp.route('/types/add', methods=("GET", "POST"))
def add():
    if request.method == 'POST':
        model = {
            'Name': request.form['name'],
            'RunAs': request.form['runAs'],
            'Registers': []
        }
        if '/' in model['Name']:
            flash('Name cannot contain forward slashes')
            return render_template('controller_types/create.html')

        _create(model)
        return redirect(
            url_for('controller_types.edit', name=request.form['name']))

    return render_template('controller_types/create.html',
                           footer=data.get_info())


@bp.route('/types/<name>/edit', methods=("GET", "POST"))
def edit(name: str):
    existing = _prepare_for_editing(name)
    if not existing:
        abort(404, f'No type with the name {name}')

    if request.method == 'POST':
        for key in request.form.keys():
            if key.startswith('add_register'):
                _add_reg(existing)
                break
            elif key.startswith('remove_'):
                _remove_reg(existing, key[7:])
                _save_model(existing, name)
                existing = _prepare_for_editing(name)
                break
            elif key.startswith('update_'):
                _update_reg(existing, key[7:], request.form)
                _save_model(existing, name)
                existing = _prepare_for_editing(name)
                break
            elif key.startswith('rename_'):
                _abort_if_cannot_delete(name)
                existing['Name'] = request.form['name']
                _save_model(existing, name)
                existing = _prepare_for_editing(existing['Name'])
                break

    return render_template('controller_types/update.html',
                           data=existing,
                           footer=data.get_info())


@bp.route('/types/<name>/copy', methods=("GET", ))
def copy(name: str):
    model = _find_model_by_name(name)
    model['Name'] = f'{model["Name"]}_copy'
    _create(model)
    return redirect(url_for('controller_types.edit', name=model['Name']))


@bp.route('/types/<name>/delete', methods=('GET', ))
def delete(name):
    models = data.get_controller_types()
    current = _abort_if_cannot_delete(name, models)
    models.remove(current)
    data.set_controller_types(models)
    return redirect(url_for("controller_types.list"))


def _find_model_by_name(name: str, models: dict = None) -> dict:
    if not models:
        models = data.get_controller_types()
    model = next(filter(lambda ct: ct['Name'] == name, models), None)
    return model


def _prepare_for_editing(name: str):
    types = ['Input', 'Holding', 'Coil', 'Discrete']
    variants = ['DropdownVariants', 'RawBinary', 'RawInteger']
    exists = _find_model_by_name(name)
    if not exists:
        return None
    for reg in exists['Registers']:
        processed = []
        for possible in reg['PossibleValues']:
            for k, v in possible.items():
                processed.append(f'{k}={v}')
        reg['PossibleValues'] = '\n'.join(processed)
    exists['RegTypes'] = types
    exists['DisplayVariants'] = variants
    return exists


def _save_model(model: dict, current_name: str):
    models = data.get_controller_types()
    models = [x for x in models if x['Name'] != current_name]
    del model['RegTypes']
    _remove_reg(model, '')
    for r in model['Registers']:
        if 'PossibleValues' in r:
            r['PossibleValues'] = _str_to_possible(r['PossibleValues'])
        else:
            r['PossibleValues'] = []
    models.insert(0, model)
    data.set_controller_types(models)


def _add_reg(model: dict):
    model['Registers'].insert(
        0, {
            'CanOverride': True,
            'Name': '',
            'Type': '',
            'Address': '',
            'PossibleValues': ''
        })


def _remove_reg(model: dict, reg_name: str):
    model['Registers'] = [
        x for x in model['Registers'] if x.get('Name') != reg_name
    ]


def _update_reg(model: dict, reg_name: str, form: dict):
    updated = {}
    for k, v in form.items():
        if k == f'{reg_name}-name':
            updated['Name'] = v
        elif k == f'{reg_name}-address':
            updated['Address'] = int(v)
        elif k == f'{reg_name}-override':
            updated['CanOverride'] = v == 'True'
        elif k == f'{reg_name}-type':
            updated['Type'] = v
        elif k == f'{reg_name}-display-as':
            updated['DisplayAs'] = v
        elif k == f'{reg_name}-possible':
            updated['PossibleValues'] = _str_to_possible(v)
    _remove_reg(model, reg_name)
    model['Registers'].insert(0, updated)


def _str_to_possible(values: str) -> list:
    if not isinstance(values, str):
        return values

    possible = []
    lines = values.splitlines()
    for line in lines:
        res = line.strip().split('=')
        possible.append({res[0]: int(res[1])})
    return possible


def _abort_if_cannot_delete(name: str, models: dict = None) -> dict:
    if not models:
        models = data.get_controller_types()
    exists = _find_model_by_name(name, models)
    if not exists:
        abort(404, f'No type with the name {name}')
    controllers = data.get_controllers()
    controller = next(filter(lambda x: x['Type'] == name, controllers), None)
    if controller:
        abort(
            409,
            f'The type {name} is in use by some controllers. Remove them first'
        )
    return exists


def _create(new_model: dict):
    models = data.get_controller_types()
    model = _find_model_by_name(new_model['Name'], models)
    if not model:
        models.append(new_model)
        data.set_controller_types(models)
